﻿-- Helperfunctions
local string_lower, string_gsub, string_format, strfind = string.lower, string.gsub, string.format, strfind
local _, wipe, tsort, tinsert = _, wipe, table.sort, table.insert
local hugeNumber, min = math.huge, math.min

-- LibDataBroker
local ldb = LibStub:GetLibrary("LibDataBroker-1.1", true)
local LDBIcon = ldb and LibStub("LibDBIcon-1.0",true)
local LibQTip = LibStub('LibQTip-1.0')
local Ailo = LibStub("AceAddon-3.0"):NewAddon("Ailo", "AceConsole-3.0", "AceEvent-3.0", "AceHook-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("Ailo", false)
local LST = LibStub("LibSyncTime-1.0", true)
local DB_VERSION = 8
local currentChar, currentRealm, currentCharRealm, currentMaxLevel, currentFaction

function Ailo:Output(...)
    if self.db.profile.showMessages then
        Ailo:Print(...)
    end
end 

-- Sorting
local sortRealms = {}
local sortRealmsPlayer = {}

local RAID_CLASS_COLORS_FONTS = {}

local function setColor(info, r, g, b, a)
    Ailo.db.profile[info[#info]] = { r = r, g = g, b = b, a = a }
end

local function getColor(info)
    return Ailo.db.profile[info[#info]].r, Ailo.db.profile[info[#info]].g, Ailo.db.profile[info[#info]].b, Ailo.db.profile[info[#info]].a
end

local function GetWeeklyResetDay()
	-- number is index into day of week; 1 == Sunday
	local resets = { 
		enCN = 3, enUS = 3, esMX = 3, ptBR = 3,                         -- Tuesday 
		deDE = 4, enGB = 4, esES = 4, frFR = 4, ptPT = 4, ruRU = 4,     -- Wednesday
		enTW = 5, koKR = 5, zhCN = 5, zhTW = 5,                         -- Thursday
	}

	local locale = GetLocale()
	local reset = resets[locale]

	if not reset then Ailo:Output(locale, "locale is not recognized, please report this error") end
	return reset or 3
end

local function GetNextDailyReset()
    -- when first logging in, we may get a bogus result from GetQuestResetTime
    -- if thats the case, then just return hugeNumber so that nothing gets purged
    local qrtime = GetQuestResetTime()
    if qrtime <= 0 then
        return hugeNumber
    end

    return Ailo:GetSyncTime() + qrtime
end

local function GetNextWeeklyReset()
    -- if we're not getting a valid number for the daily reset, then we can't
    -- calculate the weekly. just return hugeNumber so that nothing gets purged 
    local nextDailyReset = GetNextDailyReset()
    if (nextDailyReset == hugeNumber) then
        return hugeNumber
    end

    local dayOfNextReset = date("*t", nextDailyReset).wday
    local daysToWeeklyReset =  math.fmod(7 + Ailo.db.global.weeklyResetDay - dayOfNextReset, 7)
    local nextWeeklyReset = nextDailyReset + (daysToWeeklyReset * 86400) -- 86400 is number of seconds in a day (24 * 60 * 60)
    -- print(date(nil, nextWeeklyReset))

    return nextWeeklyReset
end


--=========================================================================--
-- data table definitions
--=========================================================================--

local defaults = {
    profile = {
        savedraid = { r=1, g=0, b=0, a=1 },
        freeraid  = { r=0, g=1, b=0, a=1 },
        hc  = "hc",
        nhc = "nhc",
        show5Man             = false,
        showAllChars         = false,
        showCharacterRealm   = false,
        showDailyHeroic      = true,
        showLeft             = false,
        showMessages         = true,
        showRealmHeaderLines = false,
        showRaidFinder       = true,
        showWeeklyRaid       = false,
        showTBVictory        = false,
        showWGVictory        = false,
        showDailyPVP         = false,
        showSeasonal         = true,
        useClassColors       = true,
        useCustomClassColors = true,
        instanceAbbr = {},
        minimapIcon = {
            hide = false,
            minimapPos = 220,
            radius = 80,
        },
        
    },
    global = {
        chars = {},				-- structure is = { realm = { char = { lockouts = { expire, encounter }, resets = { expire, encounter } } } }  
        charClass = {},
        raids = {},				-- structure is = { instance = { size } }
        nextPurge = hugeNumber,
        weeklyResetDay = 3,     -- TODO - set default based on region
        version = 0,
    },
}

local Seasonal = {}
Seasonal.Events = {
    Valentines = { icon = "|TInterface\\Icons\\inv_valentinesboxofchocolates02:0|t", 
                  first = time( { year = 2012, month = 2,  day = 5,  hour = 0,  min = 1  } ),
                   last = time( { year = 2012, month = 2,  day = 18, hour = 23, min = 59 } ),
             dungeon_id = 288 },
     Midsummer = { icon = "|TInterface\\Icons\\inv_misc_bag_17:0|t", 
                  first = time( { year = 2012, month = 6,  day = 21, hour = 0,  min = 1  } ),
                   last = time( { year = 2012, month = 7,  day = 4,  hour = 23, min = 59 } ),
             dungeon_id = 286 },
      Brewfest = { icon = "|TInterface\\Icons\\inv_cask_02:0|t", 
                  first = time( { year = 2012, month = 9,  day = 20, hour = 0,  min = 1  } ),
                   last = time( { year = 2012, month = 10, day = 5,  hour = 23, min = 59 } ),
             dungeon_id = 287 },
    HallowsEnd = { icon = "|TInterface\\Icons\\inv_misc_bag_28_halloween:0|t", 
                  first = time( { year = 2012, month = 10, day = 18, hour = 1,  min = 0  } ),
                   last = time( { year = 2012, month = 10, day = 31, hour = 23, min = 59 } ),
             dungeon_id = 285 },
    WinterVeil = { icon = "|TInterface\\Icons\\inv_holiday_christmas_present_01:0|t",
                  first = time( { year = 2011, month = 12, day = 15, hour = 6 } ),
                   last = time( { year = 2012, month = 1,  day = 2,  hour = 6 } ),
              quest_ids = { 6983, 7043 }, },
}

local RaidFinder = { 
    { id = 416, count = 4, expire = GetNextWeeklyReset }, -- The Siege of Wyrmrest Temple
    { id = 417, count = 4, expire = GetNextWeeklyReset }, -- Fall of Deathwing
}

local RepeatableQuests = {
    -- ID's of all raid weekly quests
    weeklydone = { expire = GetNextWeeklyReset, questIds = { 24590, 24589, 24588, 24587, 24586, 24585, 24584, 24583, 24582, 24581, 24580, 24579 } },
    -- ID's of horde and alliance versions of Victory in Wintergrasp weekly pvp quest 
    wgvictory = { expire = GetNextWeeklyReset, questIds = { 13181, 13183 } },
    -- ID's of horde and alliance versions of the Victory in Tol Barad weekly quest 
    tbvictory = { expire = GetNextWeeklyReset, questIds = { 28882, 28884 } },
    -- ID's of all horde and alliance versions of repeatable Seasonal quests
    -- Note: supports only one h/a set of quests per event, but all events can live
    -- in same struct as long as they don't have overlapping run times.
    seasonal = { expire = GetNextDailyReset, questIds = { 6983, 7043 } },
}


--=========================================================================--
-- Addon Initialization
--=========================================================================--

function Ailo:OnInitialize()

    currentMaxLevel = MAX_PLAYER_LEVEL_TABLE[GetExpansionLevel()]
    currentChar = UnitName("player")
    currentRealm = GetRealmName()
    currentCharRealm = currentChar..' - '..currentRealm
    currentFaction = string_lower(UnitFactionGroup("player"))
    defaults.global.weeklyResetDay = GetWeeklyResetDay()

    if currentMaxLevel == UnitLevel("player") then 
        self:RegisterEvent("CHAT_MSG_SYSTEM")
        self:RegisterEvent("UPDATE_INSTANCE_INFO", "UpdateLockouts")
        self:RegisterEvent("LFG_COMPLETION_REWARD")
        self:RegisterEvent("LFG_UPDATE_RANDOM_INFO", "UpdateLFGResets")
        self:RegisterEvent("QUEST_COMPLETE")
        self:SecureHook("GetQuestReward", "GetQuestRewardHook");
        self:RegisterEvent("QUEST_FINISHED")
    end

    self.db = LibStub("AceDB-3.0"):New("AiloDB", defaults, true)
    LibStub("AceConfig-3.0"):RegisterOptionsTable("Ailo", self.GenerateOptions)
    self.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("Ailo")

    if not self.db.global.version or DB_VERSION > self.db.global.version then
        self:Output(L["DB_VERSION_UPGRADE_PURGE"])
        Ailo:WipeDB()
        self.db.global.version = DB_VERSION 
    end

    local AiloLDB = ldb:NewDataObject("Ailo", {
        type = "data source",
        text = "Ailo",
        icon = "Interface\\Icons\\Achievement_Dungeon_UlduarRaid_Archway_01.png",
        OnClick = function(clickedframe, button)
            if button == "RightButton" then 
                InterfaceOptionsFrame_OpenToCategory(Ailo.optionsFrame) 
            else 
                if IsShiftKeyDown() then
                    Ailo:ManualPlayerUpdate() 
                else
                    ToggleFriendsFrame(4)
                end
            end
        end,
        OnEnter = function(tt)
            local tooltip = LibQTip:Acquire("AiloTooltip", 1, "LEFT") 
            Ailo.tooltip = tooltip
            Ailo:PrepareTooltip(tooltip) 
            tooltip:SmartAnchorTo(tt)
            tooltip:Show()
        end,
        OnLeave = function(tt)
            LibQTip:Release(Ailo.tooltip)
            Ailo.tooltip = nil
        end,
    })

    LDBIcon:Register("Ailo", AiloLDB, self.db.profile.minimapIcon)

    self:SetupClasscoloredFonts()
    if CUSTOM_CLASS_COLORS then
        CUSTOM_CLASS_COLORS:RegisterCallback("SetupClasscoloredFonts", self)
    end

    -- trigger collection of lockout, reset and quest information
    self:UpdatePlayer()
end


--=========================================================================--
--=========================================================================--
--
-- main data acquisition and processing function
--
--=========================================================================--
--=========================================================================--

function Ailo:UpdatePlayer()
    if currentMaxLevel > UnitLevel("player") then return end
    self.db.global.charClass[currentCharRealm] = select(2,UnitClass('player'))
    self:PrepareCharacterTables()
    self:UpdateQuests()

    -- trigger an UPDATE_INSTANCE_INFO event so we can collect raid/instance lockout information
    RequestRaidInfo()

    -- trigger an LFG_UPDATE_RANDOM_INFO event so we can collect information about random group rewards
    RequestLFDPlayerLockInfo()

end

--=========================================================================--
-- Raid and Instance Lockouts
--=========================================================================--

local INSTANCE_SAVED = _G["INSTANCE_SAVED"]
function Ailo:CHAT_MSG_SYSTEM(event, msg)
    if tostring(msg) == INSTANCE_SAVED then
        -- You are now saved to this instance; refresh Lockout info
        RequestRaidInfo()  -- triggers UPDATE_INSTANCE_INFO event
    end
end

-- event handler for UPDATE_INSTANCE_INFO, process all raid/instance lockouts
function Ailo:UpdateLockouts()
    if currentMaxLevel > UnitLevel("player") then return end
    self.db.global.charClass[currentCharRealm] = select(2,UnitClass('player'))

    local now = self:GetSyncTime()
    for index=1, GetNumSavedInstances() do
        local instanceName, _, instanceReset, _, locked, _, _, isRaid, maxPlayers, _, maxEncounter, encounterDone  = GetSavedInstanceInfo(index)
        if maxPlayers == 25 then 
            maxPlayers = 10 
        end
        if locked then
            local instanceString = string_format("%s.%d", instanceName, maxPlayers)
            self:ExtendRaidTable( instanceName, maxPlayers )
            self:SaveRaidForChar( instanceString, now+instanceReset, maxEncounter-encounterDone )
        end
    end
    
end

function Ailo:SaveRaidForChar(instance, expireTime, encounterLeft, character, realm)
    character     = character or currentChar
    realm         = realm or currentRealm
    encounterLeft = encounterLeft or 0

    local lockouts = self.db.global.chars[realm][character].lockouts
    lockouts[instance] = lockouts[instance] or {}
    lockouts[instance].expire    = expireTime
    lockouts[instance].encounter = encounterLeft

    if expireTime < self.db.global.nextPurge then
        self.db.global.nextPurge = expireTime
    end
end

--=========================================================================--
-- LFG Reward Resets
--=========================================================================--

function Ailo:LFG_COMPLETION_REWARD()
    --[[
    Fires when a random dungeon is completed and the achievement-like
    alert window pops up. The problem is that this DOES NOT update
    the return values of GetLFGDungeonRewards(x), those are updated
    when LFG_UPDATE_RANDOM_INFO is recieved, so force the client to
    call for an update
    ]]--
    RequestLFDPlayerLockInfo()
end

-- event handler for LFG_UPDATE_RANDOM_INFO, process all weekly/daily instance resets
function Ailo:UpdateLFGResets()
    local resets = self.db.global.chars[currentRealm][currentChar].resets
    
    local bestchoice = GetRandomDungeonBestChoice() or 301
    resets.dailyheroic = { expire = GetNextWeeklyReset(), encounter = self:GetNumCompletionsLeft( bestchoice ) }

    if GetRandomBGHonorCurrencyBonuses() then
        resets.dailypvp = { expire = GetNextDailyReset() }
    end

    local id = Seasonal:Dungeon()
    if (id and GetLFGDungeonRewards(id)) then
        resets.seasonal = { expire = GetNextDailyReset() }
    end

    for _, rf in pairs(RaidFinder) do
        local _, killed = GetLFGDungeonNumEncounters(rf.id)
        if killed > 0 then
            resets[rf.id] = { expire = rf.expire(), encounter = (rf.count - killed), done = GetLFGDungeonRewards(rf.id) }
        end
    end
end

function Ailo:GetNumCompletionsLeft(dungeonID)
    local currencyID, currencyQuantity, specificQuantity, specificLimit, overallQuantity, overallLimit, periodPurseQuantity, periodPurseLimit, isWeekly = GetLFGDungeonRewardCapInfo(dungeonID);
    if specificLimit then
        local remainingAllotment = min(specificLimit - specificQuantity, overallLimit - overallQuantity);
        if ( periodPurseLimit ~= 0 ) then
            remainingAllotment = min(remainingAllotment, periodPurseLimit - periodPurseQuantity);
        end
         
        return ceil(remainingAllotment / currencyQuantity);
    else 
        return 0
    end
end

--=========================================================================--
-- Daily/Weekly Repeatable Quests
--=========================================================================--

function Ailo:UpdateQuests()
    self:PrepareCharacterTables()
    local resets = self.db.global.chars[currentRealm][currentChar].resets

    for objective,data in pairs(RepeatableQuests) do
        resets[objective] = nil
        for _,id in pairs (data.questIds) do
            if IsQuestFlaggedCompleted(id) then
                resets[objective] = { expire = data.expire() }
                break
            end
        end
    end
end

-- After processing a QueryQuestsCompleted call, blizzard will put the user in a cooldown for that function
-- meaning it won't always return results. Usually, immediately after turning in quest Ailo would fail to 
-- update the cell to show the completed state until the server would process another QueryQuestsCompleted
-- request. Often a user could log out before this occured. The next three events/hooks should catch the act
-- of turning in a tracked quest allowing us to update UI.

-- QUEST_COMPLETE event is fired when npc is offering final dialog to turn in a quest
-- see if its one of ours and if so track it
local completedQuestId = false
function Ailo:QUEST_COMPLETE()
    completedQuestId = GetQuestID()
    for objective,data in pairs(RepeatableQuests) do
        for _,id in pairs(data.questIds) do
            if id == completedQuestId then
                return
            end
        end
    end
    
    completedQuestId = false
end

-- I believe all user actions to finalize the completion of a quest run through GetQuestReward.
-- Hook that function and if its called while we're tracking a completedQuestId, then that quest was turned in.
function Ailo:GetQuestRewardHook()
    if completedQuestId then
        for objective,data in pairs(RepeatableQuests) do
            for _,id in pairs(data.questIds) do
                if id == completedQuestId then
                    self.db.global.chars[currentRealm][currentChar].resets[objective] = { expire = data.expire() }
                    completedQuestId = false
                end
            end
        end
    end
end
    
-- QUEST_FINISHED event is fired anytime the user leaves the dialog for which a QUEST_COMPLETE event was fired,
-- whether they canceled or turned in the quest. Cleanup here by clearing the tracking variable.
function Ailo:QUEST_FINISHED()
    completedQuestId = false
end


--=========================================================================--
--=========================================================================--
--
-- saved data maintenance functions
--
--=========================================================================--
--=========================================================================--

function Ailo:PurgeExpiredData()
    local charsdb = self.db.global.chars
    local now = self:GetSyncTime()
    for realm, chars in pairs(charsdb) do 
        for char, charData in pairs(chars) do 
            if charData.lockouts then 
                for instance, instanceData in pairs(charData.lockouts) do
                    if type(instanceData) ~= "table" or not instanceData.expire or now > instanceData.expire then
                     -- self.db.global.chars[realm][char].lockouts[instance].expire    = nil
                     -- self.db.global.chars[realm][char].lockouts[instance].encounter = nil
                        self.db.global.chars[realm][char].lockouts[instance] = nil
                    end
                end
            end
            if charData.resets then 
                for instance, instanceData in pairs(charData.resets) do
                    if type(instanceData) ~= "table" or not instanceData.expire or now > instanceData.expire then
                     -- self.db.global.chars[realm][char].resets[instance].expire    = nil
                     -- self.db.global.chars[realm][char].resets[instance].encounter = nil
                        self.db.global.chars[realm][char].resets[instance] = nil
                    end
                end
            end
        end
    end
end

function Ailo:GetNextPurge()
    local charsdb = self.db.global.chars
    local nextPurge = GetNextDailyReset()
    for realm, chars in pairs(charsdb) do 
        for _, charData in pairs(chars) do 
            if charData.lockouts then
                for _, instanceData in pairs(charData.lockouts) do
                    nextPurge = min(nextPurge, instanceData.expire)
                end
            end
            if charData.resets then
                for _, instanceData in pairs(charData.resets) do
                    nextPurge = min(nextPurge, instanceData.expire)
                end
            end
        end
    end
    return nextPurge
end

function Ailo:GetSyncTime()
	return LST and LST:synctime() or time()
end

function Ailo:WipeDB()
    if type(self.db.global.chars) == "table" then 
        wipe(self.db.global.chars)
    end
    if type(self.db.global.raids) == "table" then 
        wipe(self.db.global.raids)
    end
    if type(self.db.global.charClass) == "table" then 
        wipe(self.db.global.charClass)
    end
    self.db.global.nextPurge = GetNextDailyReset()
end

function Ailo:PrepareCharacterTables()
    if not self.db.global.chars[currentRealm] then
        self.db.global.chars[currentRealm] = {}
    end
    if not self.db.global.chars[currentRealm][currentChar] then
        self.db.global.chars[currentRealm][currentChar] = {}
    end
    if not self.db.global.chars[currentRealm][currentChar].lockouts then
        self.db.global.chars[currentRealm][currentChar].lockouts = {}
    end
    if not self.db.global.chars[currentRealm][currentChar].resets then
        self.db.global.chars[currentRealm][currentChar].resets = {}
    end
end


function Ailo:ExtendRaidTable(instanceName, size)
    -- Save it for the tooltip
    self.db.global.raids[instanceName] = self.db.global.raids[instanceName] or {}
    self.db.global.raids[instanceName][size] =  true
end

function Ailo:TrimRaidTable()
    local charsdb = self.db.global.chars
    local raidsdb = self.db.global.raids
    for raid, sizes in pairs(raidsdb) do
        for size, _ in pairs(sizes) do
            local isUsed = false
            local raidString = string_format("%s.%d", raid, size)
            for realm, chars in pairs(charsdb) do 
                for char, charData in pairs(chars) do
                    if charData.lockouts and charData.lockouts[raidString] and charData.lockouts[raidString].expire then
                        isUsed = true
                        break
                    end
                end
            end
            if not isUsed then
                self:DeleteFromRaidTable( raid )
            end
        end
    end
end

function Ailo:DeleteFromRaidTable( instanceName )
    self.db.global.raids[instanceName] = nil
end


function Ailo:ManualPlayerUpdate()
    if currentMaxLevel > UnitLevel("player") then return end
    self:Output(L["Updating data for current player."])

    if not self.db.global.chars[currentRealm] then 
        self.db.global.chars[currentRealm] = {}
    else
        wipe(self.db.global.chars[currentRealm][currentChar])
        self.db.global.chars[currentRealm][currentChar] = nil
    end

    self:TrimRaidTable()
    self:UpdatePlayer()
end


function Ailo:GetInstanceAbbr(instanceName)
    if not self.db.profile.instanceAbbr[instanceName] then
        -- Has no abbreviation yet, try it with a somewhat good guess
        -- Tries to get the first char of every word, does not go well with utf-8 chars
        self.db.profile.instanceAbbr[instanceName] = string_gsub(instanceName, "(%a)[%l%p]*[%s%-]*", "%1")
    end
    
    return ( self.db.profile.instanceAbbr[instanceName] ~= "" and self.db.profile.instanceAbbr[instanceName] or nil )
end


function Ailo:SetupClasscoloredFonts()
    local class, color, CHOOSEN_CLASS_COLORS

    if self.db.profile.useCustomClassColors and CUSTOM_CLASS_COLORS then
        CHOOSEN_CLASS_COLORS = CUSTOM_CLASS_COLORS
    else
        CHOOSEN_CLASS_COLORS = RAID_CLASS_COLORS
    end
    for class,color in pairs(CHOOSEN_CLASS_COLORS) do
        if not RAID_CLASS_COLORS_FONTS[class] then 
            RAID_CLASS_COLORS_FONTS[class] = CreateFont("ClassFont"..class)
            RAID_CLASS_COLORS_FONTS[class]:CopyFontObject(GameTooltipText)
        end
        RAID_CLASS_COLORS_FONTS[class]:SetTextColor(color.r, color.g, color.b)
    end
end


function Ailo:BuildSortedKeyTables()
    local c, r
    wipe(sortRealms)
    sortRealms = {}
    wipe(sortRealmsPlayer)
    sortRealmsPlayer = {}
    for r,_ in pairs(self.db.global.chars) do
        tinsert(sortRealms, r)
        sortRealmsPlayer[r] = {}
        for c,_ in pairs(self.db.global.chars[r]) do
            tinsert(sortRealmsPlayer[r],c)
        end
        tsort(sortRealmsPlayer[r])
    end
    tsort(sortRealms)
end


-- Seasonal Quest Helper functions

function Seasonal:GetCurrentEvent()
    local now = Ailo:GetSyncTime()

    for k, event in pairs(self.Events) do
        if ( event.first <= now and now <= event.last) then
            return event
        end
    end

    return nil
end

function Seasonal:Show()
    return Ailo.db.profile.showSeasonal and self:GetCurrentEvent() ~= nil
end

function Seasonal:Icon()
    local event = self:GetCurrentEvent()
    return (event ~= nil) and event.icon or nil
end

function Seasonal:Dungeon()
    local event = self:GetCurrentEvent()
    return (event ~= nil) and event.dungeon_id or nil
end

function Seasonal:Quests()
    local event = self:GetCurrentEvent()
    return (event ~= nil) and event.quest_ids or nil
end

--=========================================================================--
--=========================================================================--
--
-- Tooltip and Options Display
--
--=========================================================================--
--=========================================================================--

--=========================================================================--
-- PrepareTooltip - build and formate the tooltip to show lockout state
--=========================================================================--

function Ailo:PrepareTooltip(tooltip)
    local raidorder = {}
    --[[[ Cell are just colored green/red
          Foo BaR Etc EtC ETC
    Char1 [ ] [ ] [ ] [ ] [ ]
    Char2 [ ] [ ] [ ] [ ] [ ]
    Char3 [ ] [ ] [ ] [ ] [ ]
    ]]--
    local charsdb = self.db.global.chars
    local raidsdb = self.db.global.raids
    local profile = self.db.profile
    
    if self:GetSyncTime() > self.db.global.nextPurge then
        -- Only search 
        self:PurgeExpiredData()
        self:TrimRaidTable()
        self.db.global.nextPurge = self:GetNextPurge()
    end

    if next(raidsdb) or profile.showAllChars or 
       Seasonal:Show() or profile.showDailyHeroic or 
       profile.showRaidFinder or profile.showWeeklyRaid or 
       profile.showWGVictory or profile.showTBVictory or profile.showDailyPVP then
        -- At least one char is saved to some 
        tooltip:AddHeader(L["Raid"]) -- Raid

        local seasonalColumn, dailyHeroicColumn, weeklyRaidColumn, dailyPVPColumn, wgVictoryColumn, tbVictoryColumn
        local raidfinder_cols = {}

        -- Seasonal column
        if Seasonal:Show() then
            seasonalColumn = tooltip:AddColumn("CENTER")
            tooltip:SetCell(1, seasonalColumn, Seasonal:Icon())
        end
        -- Daily Heroic column
        if profile.showDailyHeroic then
            dailyHeroicColumn = tooltip:AddColumn("CENTER")
            tooltip:SetCell(1, dailyHeroicColumn, "150|TInterface\\Icons\\pvecurrency-valor:0|t")
        end
        
        -- Raid Finder columns
        if profile.showRaidFinder then
            for i,rf in ipairs(RaidFinder) do
                local name = GetLFGDungeonInfo(rf.id)
                local abbr = self:GetInstanceAbbr(name)
                raidfinder_cols[i] = tooltip:AddColumn("CENTER")
                tooltip:SetCell(1, raidfinder_cols[i], abbr)
            end
        end
        
        -- Weekly Raid column
        if profile.showWeeklyRaid then
            weeklyRaidColumn = tooltip:AddColumn("CENTER")
            tooltip:SetCell(1, weeklyRaidColumn, "150|TInterface\\Icons\\pvecurrency-justice:0|t")
        end
        
        -- PvP Daily column
        if profile.showDailyPVP then
            dailyPVPColumn = tooltip:AddColumn("CENTER")
            tooltip:SetCell(1, dailyPVPColumn, "100|TInterface\\Icons\\pvpcurrency-conquest-"..currentFaction..":0|t")
        end
        -- Wintergrasp Victory column
        if profile.showWGVictory then
            wgVictoryColumn = tooltip:AddColumn("CENTER")
            tooltip:SetCell(1, wgVictoryColumn, "24|TInterface\\Icons\\pvpcurrency-honor-"..currentFaction..":0|t")
        end
        -- Tol Barad Victory column
        if profile.showTBVictory then
            tbVictoryColumn = tooltip:AddColumn("CENTER")
            tooltip:SetCell(1, tbVictoryColumn, "200|TInterface\\Icons\\pvpcurrency-honor-"..currentFaction..":0|t")
        end
        
        -- Instances with lockouts
        for raid, sizes in pairs(raidsdb) do
            local raidabbr = self:GetInstanceAbbr(raid)
            for size, _ in pairs(sizes) do
                if raidabbr and ( size > 5 or self.db.profile.show5Man ) then
                    local lastcolumn = tooltip:AddColumn("CENTER")
                    raidorder[(string_format("%s.%d", raid, size))] = lastcolumn
                    tooltip:SetCell(1, lastcolumn , raidabbr)
                end
            end
        end
        self:BuildSortedKeyTables()
        local lastline
        for _, realm in ipairs(sortRealms) do
            if profile.showRealmHeaderLines then
                lastline = tooltip:AddLine("")
                tooltip:SetCell(lastline, 1, realm, nil, "CENTER", tooltip:GetColumnCount())
            end
            for _, char in ipairs(sortRealmsPlayer[realm]) do
                local charData = charsdb[realm][char]
                if profile.showAllChars or next(charData.lockouts) or next(charData.resets) --[[ or
                   charData.weeklydone or charData.wgvictory or charData.tbvictory or charData.dailypvp --]] then

                    local charRealm = char.." - "..realm
                    local lineName = profile.showCharacterRealm and charRealm or char
                    local lineFont = profile.useClassColors and RAID_CLASS_COLORS_FONTS[ self.db.global.charClass[charRealm] ] or nil
                    lastline = tooltip:AddLine("")
                    tooltip:SetCell(lastline, 1, lineName, lineFont)

                    -- init cells
                    for i = tooltip:GetColumnCount(),2,-1 do
                        tooltip:SetCell(lastline, i, "")
                        tooltip:SetCellColor(lastline, i, profile.freeraid.r, profile.freeraid.g, profile.freeraid.b, profile.freeraid.a)
                    end

                    if seasonalColumn and charData.resets.seasonal then
                        tooltip:SetCellColor(lastline, seasonalColumn, profile.savedraid.r, profile.savedraid.g, profile.savedraid.b, profile.savedraid.a)
                    end
                    if dailyHeroicColumn and charData.resets.dailyheroic then 
                        if charData.resets.dailyheroic.encounter > 0 then
                            tooltip:SetCell(lastline, dailyHeroicColumn, charData.resets.dailyheroic.encounter)
                        else
                            tooltip:SetCellColor(lastline, dailyHeroicColumn, profile.savedraid.r, profile.savedraid.g, profile.savedraid.b, profile.savedraid.a)
                        end
                    end

                    for i, column in ipairs(raidfinder_cols) do
                        local id = RaidFinder[i].id
                        if charData.resets[id] then
                            if charData.resets[id].encounter == 0 then
                                tooltip:SetCellColor(lastline, column, profile.savedraid.r, profile.savedraid.g, profile.savedraid.b, profile.savedraid.a)
                            elseif profile.showLeft then
                                local strCell = string_format("%s%i", charData.resets[id].done and "" or "|TInterface\\Icons\\pvecurrency-valor:0|t ", charData.resets[id].encounter)
                                tooltip:SetCell(lastline, column, strCell)
                            end
                        end
                    end
                    
                    if weeklyRaidColumn and charData.resets.weeklydone then
                        tooltip:SetCellColor(lastline, weeklyRaidColumn, profile.savedraid.r, profile.savedraid.g, profile.savedraid.b, profile.savedraid.a)
                    end
                    if dailyPVPColumn and charData.resets.dailypvp then
                        tooltip:SetCellColor(lastline, dailyPVPColumn, profile.savedraid.r, profile.savedraid.g, profile.savedraid.b, profile.savedraid.a)
                    end
                    if wgVictoryColumn and charData.resets.wgvictory then
                        tooltip:SetCellColor(lastline, wgVictoryColumn, profile.savedraid.r, profile.savedraid.g, profile.savedraid.b, profile.savedraid.a)
                    end
                    if tbVictoryColumn and charData.resets.tbvictory then
                        tooltip:SetCellColor(lastline, tbVictoryColumn, profile.savedraid.r, profile.savedraid.g, profile.savedraid.b, profile.savedraid.a)
                    end
                    for instance, instanceData in pairs(charData.lockouts) do
                        if raidorder[instance] then
                            tooltip:SetCellColor(lastline, raidorder[instance], profile.savedraid.r, profile.savedraid.g, profile.savedraid.b, profile.savedraid.a)
                            if instanceData.encounter > 0 and profile.showLeft then
                                tooltip:SetCell(lastline, raidorder[instance], instanceData.encounter)
                            end
                        end
                    end
                end
            end
        end
    else
        -- No saved raids at all
        tooltip:AddHeader(L["No saved raids found"])
    end
end


--=========================================================================--
-- GenerateOptions - build options UI table
--=========================================================================--

function Ailo:GenerateOptions()
    local function newCounter()
        local i = 0
        return function()
                i = i + 1
                return i
            end
    end
    local count = newCounter()
    
    Ailo.options = {
        name = "Ailo",
        type = 'group',
        args = {
            genconfig = {
                name = L["General Settings"],
                type = 'group',
                order = 1,
                get = function(info) return Ailo.db.profile[info[#info]] end,
                set = function(info, value) Ailo.db.profile[info[#info]] = value end,
                args = {
                    savedraid = {
                        name = L["Saved raid color"],
                        desc = L["SAVED_RAID_DESC"],
                        type = 'color',
                        order = count(),
                        get  = getColor,
                        set  = setColor,
                        hasAlpha = true,
                    },
                    freeraid = {
                        name = L["Free raid color"],
                        desc = L["FREE_RAID_DESC"],
                        type = 'color',
                        order = count(),
                        get  = getColor,
                        set  = setColor,
                        hasAlpha = true,
                    },
                    useClassColors = {
                        type = "toggle",
                        order = count(),
                        name = L["Color names by class"],
                    },
                    useCustomClassColors = {
                        type = "toggle",
                        order = count(),
                        name = L["Use !ClassColors"],
                        desc = L["Use !ClassColors addon for class colors used to color the names in the tooltip"],
                        get = function(info) return Ailo.db.profile[info[#info]] end,
                        set = function(info, value) 
                            Ailo.db.profile[info[#info]] = value 
                            Ailo:SetupClasscoloredFonts()
                        end,
                        disabled = function() return not Ailo.db.profile.useClassColors or not CUSTOM_CLASS_COLORS end,
                    },
                    showCharacterRealm = {
                        name = L["Show character realms"],
                        type = "toggle",
                        order = count(),
                    },
                    minimapIcon = {
                        type = "toggle",
                        name = L["Show minimap button"],
                        desc = L["Show the Ailo minimap button"],
                        order = count(),
                        get = function(info) return not Ailo.db.profile.minimapIcon.hide end,
                        set = function(info, value)
                            Ailo.db.profile.minimapIcon.hide = not value
                            if value then LDBIcon:Show("Ailo") else LDBIcon:Hide("Ailo") end
                        end,
                    },
                    hc = {
                        type = "input",
                        order = count(),
                        disabled = true,
                        hidden = true,
                        name = L["Tooltip abbreviation used for heroic raids"],
                        width = "double",
                    },
                    nhc = {
                        type = "input",
                        order = count(),
                        disabled = true,
                        hidden = true,
                        name = L["Tooltip abbreviation used for nonheroic raids"],
                        width = "double",
                    },
                    weeklyResetDay = {
                        type = "select",
                        order = count(),
                        name = L["Weekly reset day"],
                        desc = L["Day when weekly caps and lockouts are reset"],
                        values = function() return { CalendarGetWeekdayNames() } end,
                        get = function(info) return Ailo.db.global.weeklyResetDay end,
                        set = function(info, value) Ailo.db.global.weeklyResetDay = value end,
                    },
                    show5Man = {
                        type = "toggle",
                        order = count(),
                        name = L["Show 5-man instances"],
                    },
                    showDailyHeroic = {
                        type = "toggle",
                        order = count(),
                        name = L["Track Valor Points"],
                        desc = L["TRACK_VALOR_POINTS_DESC"],
                    },
                    showRaidFinder = {
                        type = "toggle",
                        order = count(),
                        name = L["Track Raid Finder instances"],
                        desc = L["Tracks any raid done through the Raid Finder tool"],
                    },
                    showWeeklyRaid  = {
                        type = "toggle",
                        order = count(),
                        name = L["Track 'Weekly Raid'"],
                        desc = L["If the character has done the 'Weekly Raid' you get in Dalaran"],
                    },
                    showWGVictory  = {
                        type = "toggle",
                        order = count(),
                        name = L["Track 'WG Victory'"],
                        desc = L["If the character has done the 'Victory in Wintergrasp' weekly pvp quest"],
                    },
                    showTBVictory  = {
                        type = "toggle",
                        order = count(),
                        name = L["Track 'TB Victory'"],
                        desc = L["If the character has done the 'Victory in Tol Barad' weekly pvp quest"],
                    },
                    showDailyPVP  = {
                        type = "toggle",
                        order = count(),
                        name = L["Track PvP daily"],
                    },
                    showSeasonal  = {
                        type = "toggle",
                        order = count(),
                        name = L["Track Seasonal bosses"],
                        desc = L["If the character has received the reward for a seasonal boss"],
                    },
                    showLeft = {
                        type = "toggle",
                        order = -5,
                        name = L["Encounters Left"],
                        desc = L["SHOW_ENCOUTERS_LEFT_DESC"],
                    },                    
                    showRealmHeaderLines  = {
                        type = "toggle",
                        order = -4,
                        name = L["Show Realm Headers"],
                        desc = L["SHOW_REALMLINES_DESC"],
                    },                    
                    showAllChars  = {
                        type = "toggle",
                        order = -3,
                        name = L["Show all chars"],
                        desc = L["Regardles of any saved instances"],
                    },                    

                    wipeDB = {
                        type = "execute",
                        name = L["Wipe Database"],
                        order = -2,
                        confirm = true,
                        func = function() Ailo:WipeDB() end,
                    },
                    showMessages = {
                        type = "toggle",
                        order = -1,
                        name = L["Chatframe Messages"],
                    },
                },
            },
            instanceAbbr = { 
                type = 'group',
                name = L["Instance Abbreviations"],
                get = function(info) return Ailo.db.profile.instanceAbbr[info[#info]] end,
                set = function(info, value) Ailo.db.profile.instanceAbbr[info[#info]] = value end,
                args = {
                    header = {
                        type = "header",
                        order = 1,
                        name = L["Change the abbreviations used in the tooltip"]
                    },
                },
            },
        },
    }
    local instance, abbr
    for instance, abbr in pairs(Ailo.db.profile.instanceAbbr) do
        Ailo.options.args.instanceAbbr.args[instance] = {
            type = "input",
            name = instance,
        }
    end
    Ailo.options.args.profile = LibStub("AceDBOptions-3.0"):GetOptionsTable(Ailo.db)
    return Ailo.options
end
